#include <arduino_uvga.h>
#include <conio.h>
#include <kbd.h>


unsigned char state;
unsigned int timecounter;
unsigned char x, y, color, tetromino;
unsigned char playground[20][10];
unsigned char randomizer;

static char const  tetrominoes  [7][4][4] = 
{
   { {15,0,0,0}, {1,1,1,1},{15,0,0,0}, {1,1,1,1} }, // I !
   { {12,12,0,0}, {12,12,0,0},{12,12,0,0}, {12,12,0,0} }, //O
   { {4,14,0,0}, {4,12,4,0},{14,4,0,0}, {8,12,8,0}  }, // T 
   { {12,6,0,0},{4,12,8,0} ,{6,12,0,0},{8,12,4,0}  }, // S 
   { {6,12,0,0},{8,12,4,0},{12,6,0,0},  {4,12,8,0} }, // Z 
   { {8,14,0,0},{8,8,12,0} ,{14,8,0,0},{12,8,8,0}  }, // L 
   { {2,14,0,0},{12,8,8,0} ,{14,2,0,0}, {8,8,12,0} } // J 
};

void setup()   {                
   microvga_init();
   state = 0;
   randomizer = 0;
}

char doesfitscr(unsigned char x2, unsigned char y2, unsigned char rot)
{
   unsigned char i, j, tet;
   
   for (i=0;i<4;i++) 
     for (j=0;j<4;j++) 
        if (tetrominoes[tetromino/4][rot][j] & (1<<(3-i))) {
          
           if (x2 + i  >9) // x overflow
             return 0;
        }        
        
   return 1;
}


char doesfit(unsigned char x2, unsigned char y2, unsigned char rot)
{
   unsigned char i, j, tet;
   
   for (i=0;i<4;i++) 
     for (j=0;j<4;j++) 
        if (tetrominoes[tetromino/4][rot][j] & (1<<(3-i))) {
           if (y2+j > 19) // y overflow
              return 0;
           if (playground[y2+j][x2+i])
               return 0;
        }        
        
   return 1;
}

void placeit()
{
   unsigned char i, j;
   
   for (i=0;i<4;i++) 
     for (j=0;j<4;j++) 
        if (tetrominoes[tetromino/4][tetromino&3][j] & (1<<(3-i))) {
           playground[y+j][x+i] = color;
        }        
}



void drawtetromino(char visible)
{
  char i,j;

  textbackground(visible? color : 0);

  for (i=0;i<4;i++)  
   for (j=0;j<4;j++)
    if (tetrominoes[tetromino/4][tetromino%4][j] & (1<<(3-i)))
    { 
      gotoxy(30+(2*(x+i)),3+(y+j));
      _putch(' ');
      _putch(' '); 
    }
}

void newtetromino()
{
  y = 0;
  x = 3;
  color = randomizer %5;
  color++;
  tetromino = (randomizer>>2)%(7*4);
  
  if (!doesfit(x,y, tetromino%4) ) {
      state = 0; // reset game
  }
}

void repaintplayground()
{
  unsigned char i,j;
  
       for (i=0;i<20;i++) {
         gotoxy(30,3+i);
         for (j=0;j<10;j++) {
           textbackground(playground[i][j]);
           // and redraw playground
           _putch(' ');
           _putch(' ');      
         }  
       }  
}

void removebottom()
{
  unsigned char z,i,j, full, repaint;
  
  repaint = 0;
  for (z=0;z<20;z++) {
    full = 1;
    for (i=0;i<10;i++)
      if (playground[z][i] == 0)
        full = 0;
        
    if (full) {
       for (i=z;i>0;i--)
         for (j=0;j<10;j++)
           playground[i][j] = playground[i-1][j];
           
       for (j=0;j<10;j++)           
         playground[0][j] = 0;    
      repaint = 1;       
    }
    
    }
  
    if (repaint) {
      repaintplayground();
    }
  
}


void down()
{
  if (!doesfit(x,y+1, tetromino%4) ) {
    placeit();
    removebottom();
    newtetromino();
    return ;
  }
  
  drawtetromino(0);
  y++;
  
  drawtetromino(1);
}

void fall()
{
  while (1) {
  if (!doesfit(x,y+1, tetromino%4) ) {
    placeit();
    removebottom();
    newtetromino();
    return ;
  }
  
  drawtetromino(0);
  y++;
  
  drawtetromino(1);
  }
}

void left()
{
   if (x<1)
     return ;
   
  if (!doesfit(x-1,y,tetromino%4))
     return ;

     drawtetromino(0);
     x--;
     drawtetromino(1);
}


void right()
{
  // check if we still fit in...
  if (!doesfitscr(x+1,y,tetromino%4))
     return ;
     
  if (!doesfit(x+1,y,tetromino%4))
     return ;
  
  drawtetromino(0);
  x++;
  drawtetromino(1);
  
}


void rot()
{
   if (!doesfitscr(x,y,(tetromino+1)%4))
     return ;
   if (!doesfit(x,y,(tetromino+1)%4))
     return ;

   drawtetromino(0);
   tetromino = (tetromino & (~3)) + (((tetromino&3) + 1) & 3);
   drawtetromino(1);
}

void initscreen()
{
    unsigned char x,y;
    
    for (x=0;x<10;x++)
      for (y=0;y<20;y++)
        playground[y][x] = 0;
    
	textbackground(BLUE);

		clrscr();
		textcolor(BLACK);
		textbackground(CYAN);
		_cputs("This tetris runs on Arduino connected to MicroVGA-TEXT (www.MicroVGA.com)");
		clreol();

textattr(0);
 for (y=0;y<20;y++) {
   gotoxy(30, 3+y);
   for (x=0;x<20;x++)
      _putch(' ');
 }
  
}

void loop()                     
{
  
  
   if (state == 0) {
      initscreen();
      state = 1;
      timecounter = 0;
      x = 0;
      y = 0;
      color = 1;
      tetromino = 0;
   }
   
   if (state == 1) {
     
     delay(100);  
     
     if (_kbhit()) {
       int key;
       key = _getch();
       randomizer += key&0xFF;
       randomizer ^= 0x55;
       switch(key) {
           case KB_LEFT: left(); break;
           case KB_RIGHT: right(); break;
           case KB_UP: rot(); break;
           case KB_DOWN: fall(); break;
       }         
     }
     randomizer++;  
     
     timecounter++;
     
     if (timecounter == 25500)
      timecounter = 0;
      down();
   }
}